# Program for a basic GA
# This simple GA is based on the algorithm 1 proposed by Coley, David A.
# An Introduction to Genetic Algorithms for Scientists and Engineers.
# Singapore: World Scientific Publishing Co. Pte. Ltd., 1999. 227 pp.
# See page 12 of the above reference.
#
# Dr. Walter F. de Azevedo Jr.
# azevedolab.net
# Porto Alegre, April 4th 2017
#

class Simple_GA01(object):
    """A class to implement a very simple GA"""
    
    # Constructor method
    def __init__(self,pop_size,max_iter,p_cross,len_str,p_mut,div_in):
        # Import library
        import numpy as np
        
        # Define atributes
        self.pop_size = pop_size            # Population size
        self.max_iter = max_iter            # Number of iterations
        self.p_cross = p_cross              # Probability of crossover
        self.len_str = len_str              # String length
        self.p_mut = p_mut                  # Probability of mutation_Coley_algo_01
        self.div_in = div_in
        
        # Set up arrays of zeros
        self.max_array = np.zeros(self.max_iter)
        self.mean_array = np.zeros(self.max_iter)
    
    # __str__ method
    def __str__(self):
        info = "Simple_GA01 class\n"
        info += "Class to generate a model based on a very simple genetic algorithm.\n"""
        info += "It is a OOP implementation of algorithm 1 described by Coley, 1999.\n"
        info += "ReferenceÇ\n"
        info += " Coley, David A. An Introduction to Genetic Algorithms for Scientists and Engineers.\n"
        info += "Singapore: World Scientific Publishing Co. Pte. Ltd., 1999. 227 pp."
        
        return info
    
    # Definition of get_max()
    def get_max(self,x):
        """Method to return the maximum value for an array"""
    
        # Import library
        import numpy as np
    
        # Find maximum
        max1 = np.max(x)
    
        return max1

    # Definition of get_mean() f
    def get_mean(self,x):
        """Method to return the mean value for an array"""
    
        # Import library
        import numpy as np
    
        # Find mean value
        mean1 = np.mean(x)
    
        return mean1
            
    # Definition of gen_bin_strings() 
    def gen_bin_strings(self):
        """Method to generate binary strings"""
    
        # Import library
        import numpy as np
    
        # Set up empty list
        bin_str_list = []
    
        # Generate binary string
        for i in range(self.pop_size):
            bin1 = np.random.randint(2,size=self.len_str)    # Binary (2) strings of length of len_str
            bin_str0 = str(bin1)
            bin_str0 = bin_str0.replace(" ","")
            bin_str0 = bin_str0.replace("[","")
            bin_str0 = bin_str0.replace("]","")
            bin_str_list.append(bin_str0)
        
        self.current_pop = bin_str_list
            
    # Definition of bin2dec()
    def bin2dec(self,string_num):
        """Method to convert binary string to decimal integer"""
    
        # Convert binary to decinal
        num1 = str(int(string_num,2))
    
        return num1

    # Definition of bin2dec_array() 
    def bin2dec_array(self,my_pop_in):
        """Method to convert a list of binary strings to an array of integer decimals"""
    
        # Import libray
        import numpy as np
    
        # Set up an array of zeros with length of pop_in
        x = np.zeros(len(my_pop_in))
    
        # Looping through population to convert from binary to an array of decimals 
        i = 0
        for line in my_pop_in:
            aux = self.bin2dec(line)
            x[i] = aux
            i+=1
    
        return x
    
    # Define sort_population
    def sort_popupation(self,my_pop):
        """Method to sort population"""
        
        # Import library
        import numpy as np
    
        # Set up empty lists
        in_list = []
        pop_out = []
   
        # Sort array
        self.fit_sorted = np.sort(self.fit)
        
        # Reverse fitness array
        self.fit_sorted = self.fit_sorted[::-1]
    
        # Convert array to list
        my_list =  list(self.fit)
    
        # Looping through f1
        for line in self.fit_sorted:
            ind1 = my_list.index(line)
            in_list.append(ind1)
    
        # Looping through list of indeces
        for line in in_list:
            index1 = int(line)
            pop_out.append(my_pop[index1])
     
        return pop_out

    # Definition of fitness_func()
    def fitness_func(self,x1):
        """Method to calculate fitness function"""
        
        f1 = x1*x1

        return f1

    # Definition of shuffle()
    def shuffle(self,pop_in):
        """Method to shuffle population"""
    
        # Import library
        import random
    
        # Shuffle by calling random.shuffle()
        random.shuffle(pop_in)
            
        return pop_in
            
    # Definition of generate_offspring method
    def generate_offsprings(self,str1,str2):
        """Method to generate offsprings"""
    
        # Import library
        import random
    
        # Define partition locus
        locus = random.randint(2,len(str1)-1)
    
        # Generate offspring
        offspring1 = str1[0:locus] + str2[locus:]
        offspring2 = str2[0:locus] + str1[locus:]
    
        return offspring1, offspring2
    
    # Definition of update_pop
    def update_pop(self,new_pop_in):
        """Method to update population after cross-over"""
        
        # Import library
        import random
        
        # Assign population to count_pop. This population size is less than self.pop_size
        count_pop = len(new_pop_in) 
    
        # Get random number
        r_n = random.random()
    
        # Loop while to generate new population with offspring
        while count_pop < self.pop_size:    # To make sure the size of population stay the same
            
            # Test cross over prob 
            if float(self.p_cross) < r_n:                   
                
                # Loop for to generate offspring (generate_offsprings())
                for i in range(int(self.pop_size/2)):
                    o1,o2 = self.generate_offsprings(new_pop_in[i],new_pop_in[i+1])
                    new_pop_in.append(o1)
                    new_pop_in.append(o2)
                    count_pop += 2
            
            # Get random number
            r_n = random.random()
        
        # Assign new population to self.current_pop
        self.current_pop = new_pop_in
    
    # Definition of selection_Coley_algo_01() (selection operator)
    def selection_Coley_algo_01(self):
        """Method for selection operator using aklgorithm 1 from Coley 1999"""
        
        # Call bin2dec_array()
        self.x = self.bin2dec_array(self.current_pop)
    
        # Call fitness_func()
        self.fit = self.fitness_func(self.x)
        
        # Sort population
        self.current_pop = self.sort_popupation(self.current_pop)
        
        # Set up empty list for temporary population
        new_pop = []
        
        # Looping throguh population (get half of the population if self.div_in = 2)
        for i in range( int( self.pop_size/self.div_in)  ):
            new_pop.append(self.current_pop[i])
        
        # Update self.current_pop with best indivudual
        self.current_pop = new_pop
            
    # Definition of cross_over 
    def cross_over_Coley_algo_01(self):
        """Method to apply cross-over operator to a population"""
    
        # Call shuffle()
        self.current_pop = self.shuffle(self.current_pop)
        
        # Call update_pop
        self.update_pop(self.current_pop)

    # Definition of mutation_Coley_algo_01 method
    def mutation_Coley_algo_01(self):
        """Method to apply mutation_Coley_algo_01 operator"""
    
        # Import library
        import random
    
        # Determine total
        total = self.pop_size*self.len_str
    
        # Some editing list-> string
        str_pop = "".join(self.current_pop)
        str_pop = str_pop.replace("\n","")
    
        # Set up empty string
        str_pop_out = ""
    
        # Carry out mutation
        for i in range(total):
            if i == int(1/self.p_mut):
                if str_pop[i] == 1:
                    str_pop_out+="0"
                else:
                    str_pop_out+="1"
            else:
                str_pop_out+=str_pop[i]
    
        # Back to string
        # Set up empty list
        pop = []
    
        # Loop to recover population after mutation_Coley_algo_01
        for i in range(self.pop_size):
            aux = ""
            for j in range(self.len_str):
                aux += str_pop_out[i+j]
            pop.append(aux)
        
        self.current_pop = pop
    
    # Definition of show_pop()
    def show_pop(self,string2show,pop2show):
        """Method to show population"""
        
        # Show information concerning the population
        print("\n"+string2show)
        
        # Looping though the elements of the population
        for line in pop2show:
            print(line)
        
    # Definition of plot_ga_evolution() method
    def plot_ga_evolution(self,x_label,y_label,legend1,legend2,file_out):
        """Method to generate a plot of the evolution of the fitness function values for each iteraction"""
    
        # Import libraries
        import numpy as np
        import matplotlib.pyplot as plt
    
        # Generate array for x-axis
        x = np.linspace(1,self.max_iter,self.max_iter)
            
        # Generate plot
        plt.plot(x,self.max_array,label=legend1)
        plt.plot(x,self.mean_array,label=legend2)
        plt.legend(loc = "lower right")
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.grid(True)
        plt.show()
        plt.savefig(file_out)
    
    # Definition of coley_algo_01 method
    def coley_algo_01(self):
        """Method to implement algorithm 1 from Coley 1999"""
        
        # Import libraries 
        import numpy as np
        import touche as tart
        import hexadec as hx

        # General parameters for the turtles
        x_min = -240        # Minimum x coordinate in pixels
        x_max =  240        # Maximum x coordinate in pixels
        y_min = -240        # Minimum y coordinate in pixels
        y_max =  240        # Maximum y coordinate in pixels
        max_rotate = 8      # Maximum rotation angle 8*45 degrees
        conv_f = 3e-7       # Conversion factor (fitness function to dimensions)
                    
        # Call gen_bin_strings()
        self.gen_bin_strings()
        
        # Call show_pop()
        self.show_pop("\nInitial Population",self.current_pop)
        
        # Looping through the number of iteractions
        for i in range(self.max_iter):
            
            # Call selection_Coley_algo_01()
            self.selection_Coley_algo_01()
            
            # Call cross_over_Coley_algo_01() method
            self.cross_over_Coley_algo_01()
            
            # Call mutation_Coley_algo_01 method
            self.mutation_Coley_algo_01()
            
            # Call show_pop()
            self.show_pop("\nPopulation for generation "+str(i),self.current_pop)
            
            # Call bin2dec_array()
            d = self.bin2dec_array(self.current_pop)
                        
            # Call fitness_func()
            d_2 = self.fitness_func(d)
            
            # Instantiating an object of the Hexa() class and assigns it to the variable h1 
            h1 = hx.Hexa()                              # String for hexadecimal
            # Invoking the method gen_hex()
            h1.gen_hex()
            hex = h1.my_hex_string
            
            # Looping through all elements
            for j in range(len(d_2)):
                
                # Define parameters for each turtle
                x = np.random.randint(x_min,x_max)          # x coordinate in pixels
                y = np.random.randint(y_min,y_max)          # y coordinate in pixels
                length  = d_2[j]*conv_f                     # Length for the turtle
                width = d_2[j]*conv_f                       # Width for the turtle
                rotate = np.random.randint(0,max_rotate)    # Rotation angle 8*45 degrees

                # Instantiating an object of the Touche() class and assigns it to the variable t 
                t = tart.Touche(x, y, length, width, rotate, "#"+hex) 

                # Invoking the method draw()
                t.draw()
                        
            # Call get_max()
            self.max_array[i] = self.get_max(d_2)
            
            # Call get_mean()
            self.mean_array[i] = self.get_mean(d_2)
        
        # Invoking the method bye()
        t.bye()
    
        # Call show_pop()
        self.show_pop("\nFinal Population",self.current_pop)
        
def main():  

    # Set up initial values for GA
    pop_size    = 8            # Population size
    max_iter    = 46           # Number of iterations
    p_cross     = 0.8          # Probability of cross-over
    len_str     = 12           # Length of strings
    p_mut       = 0.02         # Probability of mutation_Coley_algo_01

    # Instantiating an object of the Simple_GA01() class and assigns it to the variable p1 
    p1 = Simple_GA01(pop_size,max_iter,p_cross,len_str,p_mut,2)
    
    # Invoking the plot_ga_evolution() method 
    p1.coley_algo_01()
    
    # Invoking the coley_algo_01() method
    p1.plot_ga_evolution("Generation","Fitness function, f","Maximum value of f","Mean value of f","ga01.png")
        
main()